@@ -5,7 +5,7 @@
 ;; Author: Oscar Figueiredo <oscar@xemacs.org>
 ;; Maintainer: Oscar Figueiredo <oscar@xemacs.org>
 ;; Created: Feb 1998
-;; Version: $Revision: 1.20 $
+;; Version: $Revision: 1.22 $
 ;; Keywords: help
 
 ;; This file is part of XEmacs
@@ -56,7 +56,8 @@
   (if (not (fboundp 'unless))
       (require 'cl)))
 
-(autoload 'custom-menu-create "cus-edit")
+(unless (fboundp 'custom-menu-create)
+  (autoload 'custom-menu-create "cus-edit"))
 
 (require 'eudc-vars)
 
@@ -147,11 +148,10 @@ one of the properties on the list."
 
 (defun eudc-lax-plist-get (plist prop &optional default)
   "Extract a value from a lax property list.
-
-LAX-PLIST is a lax property list, which is a list of the form (PROP1
-VALUE1 PROP2 VALUE2...), where comparison between properties is done
+PLIST is a lax property list, which is a list of the form (PROP1
+VALUE1 PROP2 VALUE2...), where comparisons between properties are done
 using `equal' instead of `eq'.  This function returns the value
-corresponding to the given PROP, or DEFAULT if PROP is not one of the
+corresponding to PROP, or DEFAULT if PROP is not one of the
 properties on the list."
   (if (not (= 0 (% (length plist) 2)))
       (error "Malformed plist"))
@@ -182,8 +182,8 @@ If PATTERN is omitted, it defaults to \"[ \\f\\t\\n\\r\\v]+\"."
 		parts)))))
 
 (defun eudc-replace-in-string (str regexp newtext)
-  "Replace all matches in STR for REGEXP with NEWTEXT string,
- and returns the new string."
+  "Replace all matches in STR for REGEXP with NEWTEXT.
+Value is the new string."
   (let ((rtn-str "")
 	(start 0)
 	match prev-start)
@@ -201,11 +201,11 @@ If PATTERN is omitted, it defaults to \"[ \\f\\t\\n\\r\\v]+\"."
 ;;{{{ Server and Protocol Variable Routines
 
 (defun eudc-server-local-variable-p (var)
-  "Return non-nil if variable has server local bindings"
+  "Return non-nil if VAR has server-local bindings."
   (eudc-plist-member (get var 'eudc-locals) 'server))
 
 (defun eudc-protocol-local-variable-p (var)
-  "Return non-nil if variable has protocol local bindings"
+  "Return non-nil if VAR has protocol-local bindings."
   (eudc-plist-member (get var 'eudc-locals) 'protocol))
 
 (defun eudc-default-set (var val)
@@ -346,9 +346,28 @@ accordingly. Otherwise it is set to its EUDC default binding"
       (setq eudc-known-protocols
 	    (cons protocol eudc-known-protocols))))
 
+;; Execute BODY restoring the server and protocol if anything goes wrong
+(defmacro eudc-server-protect (&rest body)
+  `(let ((eudc-former-server eudc-server)
+	 (eudc-former-protocol eudc-protocol))
+     (condition-case signal
+	(progn
+	  ,@body
+	  (or (and (equal eudc-server eudc-former-server)
+		   (equal eudc-protocol eudc-former-protocol))
+	      (eudc-set-server eudc-former-server eudc-former-protocol t)))
+       (t (or (and (equal eudc-server eudc-former-server)
+		   (equal eudc-protocol eudc-former-protocol))
+	      (eudc-set-server eudc-former-server eudc-former-protocol t))
+	  (signal (car signal) (cdr signal))))))
+
+(defun eudc-echo-progress (server protocol)
+  (message "Querying %s with protocol %s..." server protocol))
 
 (defun eudc-translate-query (query)
-  "Translate attribute names of QUERY according to `eudc-protocol-attributes-translation-alist'."
+  "Translate attribute names of QUERY.
+The translation is done according to 
+`eudc-protocol-attributes-translation-alist'."
   (if eudc-protocol-attributes-translation-alist
       (mapcar '(lambda (attribute)
 		 (let ((trans (assq (car attribute) 
@@ -360,7 +379,9 @@ accordingly. Otherwise it is set to its EUDC default binding"
     query)) 
 
 (defun eudc-translate-attribute-list (list)
-  "Translate a list of attribute names according to `eudc-protocol-attributes-translation-alist'."
+  "Translate a list of attribute names LIST.
+The translation is done according to
+`eudc-protocol-attributes-translation-alist'."
   (if eudc-protocol-attributes-translation-alist
       (let (trans)
 	(mapcar '(lambda (attribute)
@@ -393,29 +414,118 @@ accordingly. Otherwise it is set to its EUDC default binding"
   (setq eudc-pre-select-window-configuration nil
 	eudc-insertion-marker nil))
 
-(defun eudc-query (query &optional return-attributes no-translation)
+(defun eudc-query (query &optional return-attributes)
    "Query the current directory server with QUERY.
 QUERY is a list of cons cells (ATTR . VALUE) where ATTR is an attribute
-name and VALUE the corresponding value.  
-If NO-TRANSLATION is non nil, ATTR is translated according to 
-`eudc-protocol-attributes-translation-alist'.
+name and VALUE the corresponding value.
 RETURN-ATTRIBUTES is a list of attributes to return defaulting to 
 `eudc-default-return-attributes'."
    (unless eudc-query-function
      (error "Don't know how to perform the query"))
-   (if no-translation
-       (funcall eudc-query-function query (or return-attributes
-					      eudc-default-return-attributes))
-		
-     (funcall eudc-query-function 
-	      (eudc-translate-query query)
-	      (cond 
-	       (return-attributes
-		(eudc-translate-attribute-list return-attributes))
-	       ((listp eudc-default-return-attributes)
-		(eudc-translate-attribute-list eudc-default-return-attributes))
-	       (t
-		eudc-default-return-attributes)))))
+   (funcall eudc-query-function 
+	    (eudc-translate-query query)
+	    (cond 
+	     (return-attributes
+	      (eudc-translate-attribute-list return-attributes))
+	     ((listp eudc-default-return-attributes)
+	      (eudc-translate-attribute-list eudc-default-return-attributes))
+	     (t
+	      eudc-default-return-attributes))))
+
+(defun eudc-multi-query (queries-or-words sentinel for-inline-expansion try-all)
+  "Query one or several directory servers.
+The servers that are queried are determined by `eudc-multi-query-policy' 
+and `eudc-max-servers-to-query'.
+If FOR-INLINE-EXPANSION is non-nil, the function is called for inline
+expansion and QUERIES-OR-WORDS is expected to be a list of individual
+words which are mapped onto `eudc-inline-query-format' to build the
+directory queries. If FOR-INLINE-EXPANSION is nil, QUERIES-OR-WORDS is
+expected to be a list of individual queries. An individual query is a
+list of cons cells (ATTR . VALUE) where ATTR is an attribute name and
+VALUE is the corresponding value.
+SENTINEL is a symbol naming a sentinel function which is called
+for each new server tried.
+If TRY-ALL is non-nil then all servers are tried in order, otherwise
+search stops when at leasta match is found on a server.
+The returned value is a list of server matches. Each server match is
+itself a list of the form ((SERVER .PROTO) MATCH1 MATCH2 ...) where
+MATCHn is a matching record found on this server.  A matching record is
+a list of cons cells (ATTR . VALUE) where ATTR is an attribute name and
+VALUE is the corresponding value"
+  (if (memq eudc-multi-query-policy 
+	    '(current-server server-then-hotlist))
+      (or eudc-server
+	  (call-interactively 'eudc-set-server))
+    (or eudc-server-hotlist
+	(error "No server in the hotlist")))
+  (let (servers response found query-formats queries return-attributes)
+    ;; Prepare the list of servers to query
+    (setq servers
+	  (cond 
+	   ((eq eudc-multi-query-policy 'hotlist)
+	    eudc-server-hotlist)
+	   ((eq eudc-multi-query-policy 'server-then-hotlist)
+	    (cons (cons eudc-server eudc-protocol)
+		  (delete (cons eudc-server eudc-protocol)
+			  (copy-sequence eudc-server-hotlist))))
+	   ((eq eudc-multi-query-policy 'current-server)
+	    (list (cons eudc-server eudc-protocol)))
+	   (t
+	    (error "Wrong value for `eudc-multi-query-policy': %S"
+		   eudc-multi-query-policy))))
+    (if (and eudc-max-servers-to-query
+	     (> (length servers) eudc-max-servers-to-query))
+	(setcdr (nthcdr (1- eudc-max-servers-to-query) servers) nil))
+    
+    (eudc-server-protect
+     (catch 'found
+       ;; Loop on the servers
+       (while servers
+	 (eudc-set-server (eudc-caar servers) (eudc-cdar servers) t)
+	 (if sentinel
+	     (funcall sentinel 
+		      (eudc-caar servers) 
+		      (eudc-cdar servers)))
+	 
+	 (if (null for-inline-expansion)
+	     (setq queries queries-or-words
+		   return-attributes nil)
+	   ;; Determine which formats apply in the query-format list
+	   (setq query-formats
+		 (eudc-extract-n-word-formats eudc-inline-query-format
+					      (length queries-or-words)))
+	   ;; Build the queries
+	   (setq queries 
+		 (if query-formats
+		     (mapcar '(lambda (format)
+				(eudc-build-query-with-words-on-format
+				 queries-or-words 
+				 format))
+			     query-formats)
+		   queries-or-words))
+	   ;; Get the return attributes from the inline expansion format
+	   (setq return-attributes (cdr eudc-inline-expansion-format)))
+
+	 ;; Loop on queries
+	 (mapcar 
+	  (function 
+	   (lambda (query)
+	     ;; If the query is a simple list of strings, 
+	     ;; try to build a valid query
+	     (or (consp (car query))
+		 (setq query (eudc-build-query-with-words-on-format query)))
+	     (setq response (cons (cons (cons eudc-server eudc-protocol)
+					(eudc-query query return-attributes))
+				  response))
+	     (setq found (or found (cadr (car response))))
+	     (if (and found (null try-all))
+		 (throw 'found response))))
+	  queries)
+	 (setq servers (cdr servers)))))
+     
+     (if (null found)
+	 (error "No match")
+       response)))
 
 (defun eudc-format-attribute-name-for-display (attribute)
   "Format a directory attribute name for display.
@@ -469,63 +579,20 @@ attribute name ATTR."
     (indent-to (+ 2 column-width))
     (eudc-print-attribute-value field)))
 
-(defun eudc-display-records (records &optional raw-attr-names)
-  "Display the record list RECORDS in a formatted buffer. 
-If RAW-ATTR-NAMES is non-nil, the raw attribute names are displayed
-otherwise they are formatted according to `eudc-user-attribute-names-alist'."
+(defun eudc-setup-record-display-buffer ()
+  "Setup a buffer to display records."
   (let ((buffer (get-buffer-create "*Directory Query Results*"))
-	inhibit-read-only
-	precords
-	(width 0)
-	beg
-	first-record
-	attribute-name)
+	inhibit-read-only)
     (switch-to-buffer buffer)    
-    (setq buffer-read-only t)
-    (setq inhibit-read-only t)
+    (setq buffer-read-only t
+	  inhibit-read-only t)
     (erase-buffer)
     (insert "Directory Query Result\n")
-    (insert "======================\n\n\n")
-    (if (null records)
-	(insert "No match found.\n"
-		(if eudc-strict-return-matches
-		    "Try setting `eudc-strict-return-matches' to nil or change `eudc-default-return-attributes'.\n"
-		  ""))
-      ;; Replace field names with user names, compute max width
-      (setq precords
-	    (mapcar 
-	     (function
-	      (lambda (record)
-		(mapcar 
-		 (function
-		  (lambda (field)
-		    (setq attribute-name 
-			  (if raw-attr-names
-			      (symbol-name (car field))
-			    (eudc-format-attribute-name-for-display (car field))))
-		    (if (> (length attribute-name) width)
-			(setq width (length attribute-name)))
-		    (cons attribute-name (cdr field))))
-		 record)))
-	     records))
-      ;; Display the records
-      (setq first-record (point))
-      (mapcar 
-       (function
-	(lambda (record)
-	  (setq beg (point))
-	  ;; Map over the record fields to print the attribute/value pairs
-	  (mapcar (function 
-		   (lambda (field)
-		     (eudc-print-record-field field width))) 
-		  record)
-	  ;; Store the record internal format in some convenient place
-	  (overlay-put (make-overlay beg (point))
-		       'eudc-record
-		       (car records))
-	  (setq records (cdr records))
-	  (insert "\n")))
-       precords))
+    (insert "======================\n\n\n")))
+
+(defun eudc-close-record-display-buffer ()
+  "Insert the closing of a record display buffer."
+  (let ((inhibit-read-only t))
     (insert "\n")
     (widget-create 'push-button
 		   :notify (lambda (&rest ignore)
@@ -537,14 +604,98 @@ otherwise they are formatted according to `eudc-user-attribute-names-alist'."
 			     (kill-this-buffer))
 		   "Quit")
     (eudc-mode)
-    (widget-setup)
-    (if first-record
-	(goto-char first-record))))
+    (widget-setup)))
+ 
+(defun eudc-display-records (records &optional raw-attr-names)
+  "Display RECORDS in a formatted buffer.
+RECORDS is a server-prefixed record list as returned by `eudc-multi-query'.
+If RAW-ATTR-NAMES is non-nil, the raw attribute names are displayed
+otherwise they are formatted according to `eudc-user-attribute-names-alist'."
+  (let ((inhibit-read-only t)
+	precords
+	(width 0)
+	beg
+	attribute-name
+	num-matches
+	server)
+
+    ;; Print the number of records in the list
+    (setq num-matches (if (cdr records)
+			  (length (cdr records))
+			0))
+    (widget-insert (format "%d match%s found on %s (%s)\n\n"
+			   num-matches
+			   (if (> num-matches 1) "es" "")
+			   (caar records)
+			   (cdar records)))
+    
+    (eudc-server-protect
+     (setq server (car records))
+     ;; We need to switch servers because value display methods depend on the server
+     (eudc-set-server (car server) (cdr server) t)
+     ;; Drop server header from the record list
+     (setq records (cdr records))
+     ;; Replace field names with user names and compute max width
+     (setq precords
+	   (mapcar 
+	    (function
+	     (lambda (record)
+	       (mapcar 
+		(function
+		 (lambda (field)
+		   (setq attribute-name 
+			 (if raw-attr-names
+			     (symbol-name (car field))
+			   (eudc-format-attribute-name-for-display (car field))))
+		   (if (> (length attribute-name) width)
+		       (setq width (length attribute-name)))
+		   (cons attribute-name (cdr field))))
+		record)))
+	    records))
+     ;; Display the records
+     (mapcar 
+      (function
+       (lambda (record)
+	 (setq beg (point))
+	 ;; Map over the record fields to print the attribute/value pairs
+	 (mapcar (function 
+		  (lambda (field)
+		    (eudc-print-record-field field width))) 
+		 record)
+	 ;; Store the record internal format in some convenient place
+	 (overlay-put (make-overlay beg (point))
+		      'eudc-record
+		      (cons server (car records)))
+	 (setq records (cdr records))
+	 (insert "\n")))
+      precords))))
+  
+(defun eudc-send-mail ()
+  "Send email to the address in the current record."
+  (interactive)
+  (unless (eq major-mode 'eudc-mode)
+    (error "This command should be called in EUDC buffers"))
+  (let ((record
+         (overlay-get (car (overlays-at (point))) 'eudc-record))
+        (mail nil))
+    (unless record (error "Not on a record"))
+    ;; Since eudc-protocol-attributes-translation-alist is server-dependent we
+    ;; need to switch servers temporarily
+    (eudc-server-protect
+     (eudc-set-server (caar record) (cdar record) t)
+     (setq mail (cdr (assq 
+		      (or 
+		       (cdr (assq 'email 
+				 (symbol-value eudc-protocol-attributes-translation-alist)))
+			 'email) 
+		     (cdr record)))))
+    (if mail
+	(funcall (get mail-user-agent 'composefunc) mail)
+      (error "Cannot determine email address"))))
 
 (defun eudc-process-form ()
   "Process the query form in current buffer and display the results."
-  (let (query-alist
-	value)
+  (let (query value result inhibit-read-only)
     (if (not (and (boundp 'eudc-form-widget-list)
 		  eudc-form-widget-list))
 	(error "Not in a directory query form buffer")
@@ -552,13 +703,19 @@ otherwise they are formatted according to `eudc-user-attribute-names-alist'."
 	       (lambda (wid-field)
 		 (setq value (widget-value (cdr wid-field)))
 		 (if (not (string= value ""))
-		     (setq query-alist (cons (cons (car wid-field) value)
-					     query-alist)))))
+		     (setq query (cons (cons (car wid-field) value)
+				       query)))))
 	      eudc-form-widget-list)
       (kill-buffer (current-buffer))
-      (eudc-display-records (eudc-query query-alist) eudc-use-raw-directory-names))))
-         
-           
+      (setq result (eudc-multi-query (list query) 'eudc-echo-progress nil t))
+      ;; Display the query results
+      (eudc-setup-record-display-buffer)
+      (mapcar (function 
+	       (lambda (server-matches)
+		 (eudc-display-records server-matches
+				       eudc-use-raw-directory-names)))
+	      result)
+      (eudc-close-record-display-buffer))))
 
 (defun eudc-filter-duplicate-attributes (record)
   "Filter RECORD according to `eudc-duplicate-attribute-handling-method'."
@@ -767,8 +924,9 @@ otherwise a list of symbols is returned."
 	      entries)))
     (error "The %s protocol has no support for listing attributes" eudc-protocol)))
 
-(defun eudc-format-query (words format)
-  "Use FORMAT to build a EUDC query from WORDS"
+(defun eudc-build-query-with-words-on-format (words &optional format)
+  "Use FORMAT to build a EUDC query from WORDS.
+If FORMAT id nil, use the protocol default attribute name, if any."
   (let (query
 	query-alist
 	key val cell)
@@ -797,7 +955,7 @@ otherwise a list of symbols is returned."
 
 (defun eudc-extract-n-word-formats (format-list n)
   "Extract a list of N-long formats from FORMAT-LIST.
-If none try N-1 and so forth."
+If none try N - 1 and so forth."
   (let (formats)
     (while (and (null formats)
 		(> n 0))
@@ -825,15 +983,9 @@ After querying the server for the given string, the expansion specified by
 `eudc-inline-expansion-format' is inserted in the buffer at point.
 If REPLACE is non nil, then this expansion replaces the name in the buffer.
 `eudc-expansion-overwrites-query' being non nil inverts the meaning of REPLACE.
-Multiple servers can be tried with the same query until one finds a match, 
-see `eudc-inline-expansion-servers'"
+Multiple servers can be tried with the same query until a match is found, 
+see `eudc-multi-query-policy'"
   (interactive)
-  (if (memq eudc-inline-expansion-servers 
-	    '(current-server server-then-hotlist))
-      (or eudc-server
-	  (call-interactively 'eudc-set-server))
-    (or eudc-server-hotlist
-	(error "No server in the hotlist")))
   (let* ((end (point))
 	 (beg (save-excursion
 		(if (re-search-backward "\\([:,]\\|^\\)[ \t]*" 
@@ -844,111 +996,57 @@ see `eudc-inline-expansion-servers'"
 		    (goto-char (match-end 0)))
 		(point)))
 	 (query-words (split-string (buffer-substring beg end) "[ \t]+"))
-	 query-formats
 	 response
 	 response-string
-	 response-strings
-	 (eudc-former-server eudc-server)
-	 (eudc-former-protocol eudc-protocol)
-	 servers)
-
-    ;; Prepare the list of servers to query
-    (setq servers (copy-sequence eudc-server-hotlist))
-    (setq servers
-	  (cond 
-	   ((eq eudc-inline-expansion-servers 'hotlist)
-	    eudc-server-hotlist)
-	   ((eq eudc-inline-expansion-servers 'server-then-hotlist)
-	    (cons (cons eudc-server eudc-protocol)
-		  (delete (cons eudc-server eudc-protocol) servers)))
-	   ((eq eudc-inline-expansion-servers 'current-server)
-	    (list (cons eudc-server eudc-protocol)))
-	   (t
-	    (error "Wrong value for `eudc-inline-expansion-servers': %S"
-		   eudc-inline-expansion-servers))))
-    (if (and eudc-max-servers-to-query
-	     (> (length servers) eudc-max-servers-to-query))
-	(setcdr (nthcdr (1- eudc-max-servers-to-query) servers) nil))
-
-    (condition-case signal
-	(progn
-	  (setq response 
-		(catch 'found
-		  ;; Loop on the servers
-		  (while servers
-		    (eudc-set-server (eudc-caar servers) (eudc-cdar servers) t)
-		    
-		    ;; Determine which formats apply in the query-format list
-		    (setq query-formats
-			  (or 
-			   (eudc-extract-n-word-formats eudc-inline-query-format
-							(length query-words))
-			   (if (null eudc-protocol-has-default-query-attributes)
-			       '(name))))
-		    
-		    ;; Loop on query-formats
-		    (while query-formats
-		      (setq response
-			    (eudc-query
-			     (eudc-format-query query-words (car query-formats))
-			     (eudc-translate-attribute-list
-			      (cdr eudc-inline-expansion-format))))
-		      (if response
-			  (throw 'found response))
-		      (setq query-formats (cdr query-formats)))
-		    (setq servers (cdr servers)))
-		  ;; No more servers to try... no match found
-		  nil))
-
-
-	  (if (null response)
-	      (error "No match")
+	 response-strings)
+				
+    (setq response
+	  (car (eudc-multi-query query-words nil t nil)))
 	    
-	    ;; Process response through eudc-inline-expansion-format
-	    (while response
-	      (setq response-string (apply 'format 
-					   (car eudc-inline-expansion-format)
-					   (mapcar (function 
-						    (lambda (field)
-						      (or (cdr (assq field (car response))) 
-							  "")))
-						   (eudc-translate-attribute-list
-						    (cdr eudc-inline-expansion-format)))))
-	      (if (> (length response-string) 0)
-		  (setq response-strings
-			(cons response-string response-strings)))
-	      (setq response (cdr response)))
-	    
-	    (if (or
-		 (and replace (not eudc-expansion-overwrites-query))
-		 (and (not replace) eudc-expansion-overwrites-query))
-		(delete-region beg end))
-	    (cond 
-	     ((or (= (length response-strings) 1)
-		  (null eudc-multiple-match-handling-method)
-		  (eq eudc-multiple-match-handling-method 'first))
-	      (insert (car response-strings)))
-	     ((eq eudc-multiple-match-handling-method 'select)
-	      (eudc-select response-strings))
-	     ((eq eudc-multiple-match-handling-method 'all)
-	      (insert (mapconcat 'identity response-strings ", ")))
-	     ((eq eudc-multiple-match-handling-method 'abort)
-	      (error "There is more than one match for the query"))
-	     ))
-	  (or (and (equal eudc-server eudc-former-server)
-		   (equal eudc-protocol eudc-former-protocol))
-	      (eudc-set-server eudc-former-server eudc-former-protocol t)))
-      (t
-       (or (and (equal eudc-server eudc-former-server)
-		(equal eudc-protocol eudc-former-protocol))
-	   (eudc-set-server eudc-former-server eudc-former-protocol t))
-       (signal (car signal) (cdr signal))))))
-  
+    ;; Process response through eudc-inline-expansion-format. We need
+    ;; to switch to the context of the server where a match was found
+    (if response
+	(eudc-server-protect
+	  ;; Switch to the context of the server where a match was found
+	 (eudc-set-server (caar response) (cdar response) t)
+	 (setq response (cdr response))	; Drop the server/protocol part
+
+	 ;; Loop over the matches found on the server
+	 (while response
+	   (setq response-string (apply 'format 
+					(car eudc-inline-expansion-format)
+					(mapcar (function 
+						 (lambda (field)
+						   (or (cdr (assq field (car response))) 
+						       "")))
+						(eudc-translate-attribute-list
+						 (cdr eudc-inline-expansion-format)))))
+	   (if (> (length response-string) 0)
+	       (setq response-strings
+		     (cons response-string response-strings)))
+	   (setq response (cdr response)))
+    
+	 (if (or
+	      (and replace (not eudc-expansion-overwrites-query))
+	      (and (not replace) eudc-expansion-overwrites-query))
+	     (delete-region beg end))
+	 (cond 
+	  ((or (= (length response-strings) 1)
+	       (null eudc-multiple-match-handling-method)
+	       (eq eudc-multiple-match-handling-method 'first))
+	   (insert (car response-strings)))
+	  ((eq eudc-multiple-match-handling-method 'select)
+	   (eudc-select response-strings))
+	  ((eq eudc-multiple-match-handling-method 'all)
+	   (insert (mapconcat 'identity response-strings ", ")))
+	  ((eq eudc-multiple-match-handling-method 'abort)
+	   (error "There is more than one match for the query")))))))
+
 ;;;###autoload
 (defun eudc-query-form (&optional get-fields-from-server)
-  "Display a form to query the directory server.
-If given a non-nil argument the function first queries the server 
-for the existing fields and displays a corresponding form."
+  "Display a form to query directory servers.
+If given a non-nil argument GET-FIELDS-FROM-SERVER, the function first
+queries the server for the existing fields and displays a corresponding form."
   (interactive "P")
   (let ((fields (or (and get-fields-from-server
 			 (eudc-get-attribute-list))
@@ -966,12 +1064,31 @@ for the existing fields and displays a corresponding form."
     (make-local-variable 'eudc-form-widget-list)
     (widget-insert "Directory Query Form\n")
     (widget-insert "====================\n\n")
-    (widget-insert "Current server is: " (or eudc-server
+    (widget-insert "Multi-query policy: "
+		   (cond
+		    ((eq eudc-multi-query-policy 'current-server)
+		     "current server only\n")
+		    ((eq eudc-multi-query-policy 'hotlist)
+		     (if (or (null eudc-max-servers-to-query)
+			     (>= eudc-max-servers-to-query
+				 (length eudc-server-hotlist)))
+			 "hotlist servers\n"
+		       (concat "First " 
+			       eudc-max-servers-to-query
+			       " servers of the hotlist\n")))
+		    ((eq eudc-multi-query-policy 'server-then-hotlist)
+		     (if (or (null eudc-max-servers-to-query)
+			     (> eudc-max-servers-to-query
+				(length eudc-server-hotlist)))
+			 "current server then hotlist\n"
+		       (format "current server then first %d servers of the hotlist\n"
+			       (1- eudc-max-servers-to-query))))))
+    (widget-insert "Current server is : " (or eudc-server
 					     (progn 
 					       (call-interactively 'eudc-set-server)
 					       eudc-server))
 					     "\n")
-    (widget-insert "Protocol         : " (symbol-name eudc-protocol) "\n")
+    (widget-insert "Protocol          : " (symbol-name eudc-protocol) "\n")
     ;; Build the list of prompts
     (setq prompts (if eudc-use-raw-directory-names
 		      (mapcar 'symbol-name (eudc-translate-attribute-list fields))
@@ -1026,7 +1143,7 @@ for the existing fields and displays a corresponding form."
   )
 
 (defun eudc-bookmark-server (server protocol)
-  "Add SERVER to the EUDC `servers' hotlist."
+  "Add SERVER using PROTOCOL to the EUDC `servers' hotlist."
   (interactive "sDirectory server: \nsProtocol: ")
   (if (member (cons server protocol) eudc-server-hotlist)
       (error "%s:%s is already in the hotlist" protocol server)
@@ -1136,9 +1253,12 @@ for the existing fields and displays a corresponding form."
 (defconst eudc-custom-generated-menu (cdr (custom-menu-create 'eudc)))
 
 (defconst eudc-tail-menu 
-  `(["---" nil nil]
+  `("---"
     ["Query with Form" eudc-query-form t]
     ["Expand Inline Query" eudc-expand-inline t]
+    ["Send mail to this address" eudc-send-mail
+     (and (overlays-at (point))
+	  (overlay-get (car (overlays-at (point))) 'eudc-record))]
     ["Insert Record into BBDB" eudc-insert-record-at-point-into-bbdb 
      (and (or (featurep 'bbdb)
 	      (prog1 (locate-library "bbdb") (message "")))
@@ -1148,16 +1268,16 @@ for the existing fields and displays a corresponding form."
      (and (eq major-mode 'eudc-mode)
 	  (or (featurep 'bbdb)
 	      (prog1 (locate-library "bbdb") (message ""))))]
-    ["---" nil nil]
+    "---"
     ["Get Email" eudc-get-email t]
     ["Get Phone" eudc-get-phone t]
     ["List Valid Attribute Names" eudc-get-attribute-list t]
-    ["---" nil nil]
+    "---"
     ,(cons "Customize" eudc-custom-generated-menu)))
     
 
 (defconst eudc-server-menu 
-  '(["---" nil nil]
+  '("---"
     ["Bookmark Current Server" eudc-bookmark-current-server t]
     ["Edit Server List" eudc-edit-hotlist t]
     ["New Server" eudc-set-server t]))
@@ -1240,32 +1360,58 @@ This does nothing except loading eudc by autoload side-effect."
   (interactive)
   nil)
 
+;;}}}
+
 ;;;###autoload
-(let ((menu  '("Directory Search"
-	       ["Load Hotlist of Servers" eudc-load-eudc t]
-	       ["New Server" eudc-set-server t]
-	       ["---" nil nil]
-	       ["Query with Form" eudc-query-form t]
-	       ["Expand Inline Query" eudc-expand-inline t]
-	       ["---" nil nil]
-	       ["Get Email" eudc-get-email t]
-	       ["Get Phone" eudc-get-phone t])))
-  (if (not (featurep 'eudc-autoloads))
-      (if (string-match "XEmacs" emacs-version)
-	  (if (and (featurep 'menubar)
-		   (not (featurep 'infodock)))
-	      (add-submenu '("Tools") menu))
-	(require 'easymenu)
-	(cond 
-	 ((fboundp 'easy-menu-add-item)
-	  (easy-menu-add-item nil '("tools") (easy-menu-create-menu (car menu)
-								    (cdr menu))))
-	 ((fboundp 'easy-menu-create-keymaps)
-	  (define-key 
-	    global-map
-	    [menu-bar tools eudc] 
-	    (cons "Directory Search"
-		  (easy-menu-create-keymaps "Directory Search" (cdr menu)))))))))
+(cond ((not (string-match "XEmacs" emacs-version))
+       (defvar eudc-tools-menu (make-sparse-keymap "Directory Search"))
+       (fset 'eudc-tools-menu (symbol-value 'eudc-tools-menu))
+       
+       (define-key eudc-tools-menu [phone]
+	 '("Get Phone" . eudc-get-phone))
+       (define-key eudc-tools-menu [email]
+	 '("Get Email" . eudc-get-email))
+       (define-key eudc-tools-menu [separator-eudc-email]
+	 '("---"))
+       (define-key eudc-tools-menu [expand-inline]
+	 '("Expand Inline Query" . eudc-expand-inline))
+       (define-key eudc-tools-menu [query]
+	 '("Query with Form" . eudc-query-form))
+       (define-key eudc-tools-menu [separator-eudc-query]
+	 '("---"))
+       (define-key eudc-tools-menu [new]
+	 '("New Server" . eudc-set-server))
+       (define-key eudc-tools-menu [load]
+	 '("Load Hotlist of Servers" . eudc-load-eudc)))
+      
+      (t
+       (let ((menu  '("Directory Search"
+		      ["Load Hotlist of Servers" eudc-load-eudc t]
+		      ["New Server" eudc-set-server t]
+		      "---"
+		      ["Query with Form" eudc-query-form t]
+		      ["Expand Inline Query" eudc-expand-inline t]
+		      "---"
+		      ["Get Email" eudc-get-email t]
+		      ["Get Phone" eudc-get-phone t])))
+	 (if (not (featurep 'eudc-autoloads))
+	     (if (string-match "XEmacs" emacs-version)
+		 (if (and (featurep 'menubar)
+			  (not (featurep 'infodock)))
+		     (add-submenu '("Tools") menu))
+	       (require 'easymenu)
+	       (cond 
+		((fboundp 'easy-menu-add-item)
+		 (easy-menu-add-item nil '("tools")
+				     (easy-menu-create-menu (car menu)
+							    (cdr menu))))
+		((fboundp 'easy-menu-create-keymaps)
+		 (define-key 
+		   global-map
+		   [menu-bar tools eudc] 
+		   (cons "Directory Search"
+			 (easy-menu-create-keymaps "Directory Search"
+						   (cdr menu)))))))))))
         
 ;;}}}
 
